
#include <functional>
#include <iostream>
#include <string>

using std::cout;
using std::endl;

struct PrintName {
  std::string name;
  PrintName(std::string n) : name(std::move(n)) {}

  void cprint() const { std::cout << name << endl; }

  void print() { cprint(); }

  std::function<void()> make_printer() {
    // Copies pointer to this object, does not copy object.  Generally
    // what you want, but returned std::function becomes invalid once
    // this PrintName object is destroyed.
    return [this](){ print(); };

    // As of C++17, you can name and capture arbitrary expressions.
    return [pnp=this](){ pnp->print(); };

    // If you capture everything by reference, also captures this.
    return [&](){ print(); };

    // This works, but causes a warning as of C++20 because it
    // confusingly copies everything referenced except current object.
    return [=](){ print(); };

    // Valid as of C++20.  Explicitly declares intent to capture
    // pointer this and maintain object in scope for life of lambda.
    //return [=,this](){ print(); };

    // Capture current object by value.  Will not reflect subsequent
    // changes to name.  Can't call non-const print() method, because
    // copy inside lambda is const.
    return [*this]() { cprint(); };

    // Sometimes you want a lambda to be able to change the value of
    // captured variables, so you can declare a mutable lambda.
    return [*this]() mutable { print(); };

    // Not a valid C++ expression:
    // return this->print;
  }
};

void
doit(std::function<void()> fn)
{
  fn();
}

int
main()
{
  PrintName pn("CS111");
  doit(pn.make_printer());

  std::function<void()> crashme;
  {
    PrintName pn2("pn2");
    crashme = pn2.make_printer();
    pn2.name = "valid while pn2 exists";
    crashme();
  }
  //doit(crashme);
}
