
#include <fcntl.h>
#include <unistd.h>

#include <cstdio>
#include <iostream>
#include <functional>
#include <utility>

int
verbose_close(int fd)
{
  if (fd >= 0) {
    std::cout << "Yes I remembered to close(" << fd << ");\n";
    return close(fd);
  }
  std::cout << "Nothing to close, fd is already -1\n";
  return 0;
}

// RAII file descriptor
struct safe_fd {
  int fd_ = -1;

  safe_fd(int fd) : fd_(fd) {}
  ~safe_fd() { verbose_close(fd_); }

  // Allow fd to be move constructed (not copied)
  safe_fd(safe_fd &&other) { std::swap(fd_, other.fd_); }
  safe_fd(const safe_fd &) = delete; // redundant when you have move constructor

  // So you can use safe_fd wherever you need a file descriptor
  operator int() const { return fd_; }

  friend int close(safe_fd &fd); // redundant when no private members
};

// Do the right thing when closing a safe_fd
int
close(safe_fd &fd)
{
  std::cout << "** in close(safe_fd &)\n";
  int result = verbose_close(fd.fd_);
  fd.fd_ = -1;
  return result;
}

safe_fd
safe_open_1(const char *path, int flags, int mode = 0666)
{
  safe_fd fd = open(path, flags, mode);
  if (fd == -1)
    std::perror(path);
  return fd;
}

void
raii()
{
  std::cout << "* raii\n";

  safe_fd fd = safe_open_1("/dev/null", O_RDONLY);

  for (;;) {
    char buf[8192];
    int n = read(fd, buf, sizeof(buf));
    if (n == 0)
      break;
    else if (n == -1) {
      std::perror("read");
      // notice we don't close
      return;
    }
  }
  close(fd);
}

class Finally : std::function<void()> {
public:
  using super = std::function<void()>;

  // Import constructors from supertype
  using super::super;
  ~Finally() { if (*this) (*this)(); }

  // Allow moving
  Finally(Finally &&other) : super(std::move(other)) { other = nullptr; }
  Finally &operator=(Finally &&other) {
    super::operator=(static_cast<super&&>(other));
    other.super::operator=(nullptr);
    return *this;
  }
};

void
generic_cleanup()
{
  std::cout << "* generic_cleanup\n";
  int fd = open("/dev/null", O_RDONLY);
  Finally cleanup([fd](){ verbose_close(fd); });

  for (;;) {
    char buf[8192];
    int n = read(fd, buf, sizeof(buf));
    if (n == 0)
      return;
    else if (n == -1) {
      std::perror("read");
      return;
    }
  }
}

std::pair<int, Finally>
safe_open_2(const char *path, int flags, int mode = 0666)
{
  if (int fd = open(path, flags, mode); fd >= 0) {
    Finally cleanup([fd](){ verbose_close(fd); });
    return {fd, std::move(cleanup)};
    // Note this was for demonstration.  Could also construct return value:
    //   return {fd, [fd](){ verbose_close(fd); }};
  }
  return {-1, nullptr};
}

void
use_safe_open_2()
{
  std::cout << "* use_safe_open_2\n";
  auto [fd, cleanup] = safe_open_2("/dev/null", O_RDONLY);

  for (;;) {
    char buf[8192];
    int n = read(fd, buf, sizeof(buf));
    if (n == 0)
      return;
    else if (n == -1) {
      std::perror("read");
      return;
    }
  }
}

int
main()
{
  raii();
  //generic_cleanup();
  //use_safe_open_2();
}
